import { Option } from './option.js'
import { thunk } from './fn.js'
/** @import { Morphism } from './types.js' */


/**
 * @template A
 * @typedef {Pure<A> | Impure<A>} Free
 */

/** @template A */
export class Pure {
  #value

  /**
   * @param {A} value 
   */
  constructor(value) {
    this.#value = value
  }

  /**
   * @template {Free<B>} B
   * @param {Morphism<A, B>} f
   * @returns {B}
   */
  chain(f) {
    console.log('Pure.chain', this.#value)
    return f(this.#value)
  }

  /**
   * @template B
   * @param {Morphism<A, B>} f
   * @returns {Free<B>}
   */
  map(f) {
    console.log(`Pure.map ${f} ${this.#value}`)
    // @ts-ignore
    return pure(f(this.#value))
  }

  /**
   * @template B, C
   * @param {Free<B>} b 
   * @returns {Free<C>}
   */
  ap(b) {
    console.log('Pure.ap', b, this.#value)
    return b.map(this.#value)
  }

  traverse(f, A) {
    return A.of(this.map(f))
  }

  /**
   * @template B
   * @param {(acc: B, value: A) => B} f 
   * @param {B} init
   * @returns {B}
   */
  reduce(f, init) {
    console.log('Pure.reduce', init, this.#value)
    return f(init, this.#value)
  }

}

/**
 * @template A
 * @typedef {<B>(value: A) => Free<B>} Computation
 */

/** @template A */
export class Impure {
  #next

  /**
   * @param {() => Free<any>} next 
   */
  constructor(next) {
    this.#next = next
  }

  /**
   * @template {Free<B>} B
   * @param {Morphism<A, B>} f
   * @returns {B}
   */
  chain(f) {
    console.log('Impure.chain')
    // @ts-ignore
    return new Impure(
      // @ts-ignore
      () => {
        console.log(`Impure.chain<${f}>`, this.#next())
        this.#next().chain(f)
      }
      //() => this.#next().chain(f)
    )
  }

  /**
   * @template B
   * @param {Morphism<A, B>} f
   * @returns {Free<B>}
   */
  map(f) {
    console.log(`Impure.map ${f}`)
    return new Impure(
      // @ts-ignore
      () => {
        console.log(`Impure.map<${f}>`, this.#next())
        return this.#next().map(f)
      }
      //() => this.#next().map(f)
    )
  }

  /**
   * @template B
   * @param {Free<any>} b 
   * @returns {Free<B>}
   */
  ap(b) {
    console.log('Impure.ap', b)
    return new Impure(
      // @ts-ignore
      () => b.map(this.#next())
    )
  }

  traverse(f, A) {
    return A.of(liftF(
      this.#next().traverse(f, A)
    ))
  }

  /**
   * @template B
   * @param {(acc: B, value: A) => B} f 
   * @param {B} acc
   * @returns {B}
   */
  reduce(f, acc) {
  }
}

/**
 * @template A
 * @param {A} x
 * @returns {Free<A, A>}
 */
export const pure = x => new Pure(x)

/**
 * @template A, B
 * @param {Computation<A, B>} f
 */
export const impure = f => new Impure(f)
export const liftF = effect => impure(thunk(effect))

